/*
    SmartHomeGeeks.io Smart Window Ventilation
    https://smarthomegeeks.io/smart-window-ventilation/
*/

#include <ESP8266WiFi.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <PubSubClient.h>
#include <ArduinoOTA.h>
#include <ArduinoJson.h>
#include <arduino_secrets.h>
#include <EEPROM.h>

// Device configuration
const int closePort = 4;                                        // close IO port
const int openPort = 5;                                         // open IO port
const int closeTime = 5500;                                     // close time in milliseconds
const int openTime = closeTime;                                 // open time in milliseconds
const String deviceName = "Window Ventilation";                 // device name

// WiFi configuration
const char* ssid = SECRET_SSID;
const char* password = SECRET_PASS;

// MQTT configuration
const String mqttClient = "window_ventilation";              // MQTT client name
const String mqttStateTopic = "window_ventilation/state";    // MQTT state topic
const String mqttCmdTopic = "window_ventilation/command";    // MQTT command topic
const String mqttAvlTopic = mqttClient + "/availability";   // MQTT availability topic
const char* mqttServer = SECRET_MQTT_BROKER;                // MQTT broker IP
const char* mqttUser = SECRET_MQTT_USERNAME;                // MQTT username
const char* mqttPassword = SECRET_MQTT_PASSWORD;            // MQTT password
const int mqttPort = 1883;                                  // MQTT port (default: 1883)

bool open;
unsigned long lastMQTTconnect = 0;

WiFiClient espClient;
PubSubClient client(espClient);

void autoDiscovery() {
  Serial.println("Publishing Home Assistant Autodiscovery...");

  char buffer[512];

  DynamicJsonDocument deviceData(512);
  deviceData["name"] = deviceName;
  deviceData["mdl"] = "Smart Window Ventilation";
  deviceData["ids"] = mqttClient;
  deviceData["mf"] = "SmartHomeGeeks.io";
  deviceData["sw"] = "2023.08";

  // Control
  DynamicJsonDocument root(512);
  root["name"] = nullptr;
  root["obj_id"] = mqttClient;
  root["ic"] = "mdi:air-filter";
  root["stat_t"] = mqttStateTopic;
  root["cmd_t"] = mqttCmdTopic;
  root["pl_stop"] = nullptr;
  root["uniq_id"] = "cover." + mqttClient;
  root["avty_t"] = mqttAvlTopic;
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/cover/" + mqttClient + "/config").c_str(), buffer, true);
  root.clear();

  // RSSI sensor
  root["name"] = "RSSI";
  root["obj_id"] = mqttClient + "_rssi";
  root["stat_t"] = mqttClient + "/rssi";
  root["unit_of_meas"] = "dBm";
  root["ic"] = "mdi:access-point";
  root["uniq_id"] = "sensor." + mqttClient + "_rssi";
  root["avty_t"] = mqttAvlTopic;
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/rssi/config").c_str(), buffer, true);
  root.clear();

  // IP
  root["name"] = "IP";
  root["obj_id"] = mqttClient + "_ip";
  root["stat_t"] = mqttClient + "/ip";
  root["ic"] = "mdi:ip-network";
  root["uniq_id"] = "sensor." + mqttClient + "_ip";
  root["avty_t"] = mqttClient + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/ip/config").c_str(), buffer, true);
  root.clear();
}

void setup_wifi() {
  delay(1000);

  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.print(ssid);

  WiFi.mode(WIFI_STA);
  WiFi.hostname(mqttClient.c_str());
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  Serial.println();
  Serial.println("WiFi connected");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
}

void setup() {
  pinMode(openPort, OUTPUT);
  pinMode(closePort, OUTPUT);

  Serial.begin(115200);

  setup_wifi();

  // Configure OTA
  ArduinoOTA.setHostname(mqttClient.c_str());
  ArduinoOTA.setPassword((const char*)SECRET_OTA_PASSWORD);
  ArduinoOTA.begin();

  // Configure MQTT
  client.setServer(mqttServer, mqttPort);
  client.setCallback(callback);
  connectmqtt();

  EEPROM.begin(1);
  open = EEPROM.read(0);

  Serial.print("Current value of open: ");
  Serial.println(open);

  client.publish(mqttStateTopic.c_str(), (open ? "open" : "closed"), true);
}

void saveOpenState() {
  Serial.print("Saving open value to EEPROM: ");
  Serial.println(open);
  EEPROM.write(0, open);
  EEPROM.commit();
}

void openVentilation() {
  Serial.println("Opening ventilation");
  client.publish(mqttStateTopic.c_str(), "opening", true);
  digitalWrite(openPort, HIGH);
  delay(openTime);
  digitalWrite(openPort, LOW);
  open = true;
}

void closeVentilation() {
  Serial.println("Closing ventilation");
  client.publish(mqttStateTopic.c_str(), "closing", true);
  digitalWrite(closePort, HIGH);
  delay(closeTime);
  digitalWrite(closePort, LOW);
  open = false;
}

void connectmqtt() {
  // prevent constant reconnect
  if (millis() - lastMQTTconnect < 5000) return;
  lastMQTTconnect = millis();

  if (client.connect(mqttClient.c_str(), mqttUser, mqttPassword, mqttAvlTopic.c_str(), 1, true, "offline")) {
    Serial.println("connected to MQTT");
    client.setBufferSize(512);
    autoDiscovery();

    // ... and resubscribe
    client.subscribe(mqttCmdTopic.c_str());
    client.publish(mqttAvlTopic.c_str(), "online", true);
    client.publish((mqttClient + "/ip").c_str(),  WiFi.localIP().toString().c_str(), true);
    client.publish((mqttClient + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
  } else {
    Serial.print("failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
  }
}

void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("Message arrived: ");
  for (int i=0;i<length;i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();
  Serial.print("Current open state: ");
  Serial.println(open);

  if (!strncmp((char*)payload, "OPEN", length) && !open) {
    openVentilation();
  } else if (!strncmp((char*)payload, "CLOSE", length) && open) {
    closeVentilation();
  }

  client.publish(mqttStateTopic.c_str(), (open ? "open" : "closed"), true);
  client.publish((mqttClient + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
  saveOpenState();
}

void loop() {
  ArduinoOTA.handle();

  if (!client.connected()) {
    connectmqtt();
  }

  client.loop();
}
